﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using Emgu.CV;
using Emgu.CV.Structure;
using bbiwarg.Images;
using bbiwarg.Utility;

namespace bbiwarg.Detectors.Fingers
{
    class FingerDetector
    {
        private DepthImage depthImage;
        private EdgeImage edgeImage;
        private FingerImage fingerImage;
        private Image<Gray, byte> fingerPointsImage;
        private List<Line2D> fingerLines;
        private List<Finger> fingers;


        public FingerDetector(DepthImage depthImage, EdgeImage edgeImage, FingerImage fingerImage) {
            this.depthImage = depthImage;
            this.edgeImage = edgeImage;
            this.fingerImage = fingerImage;

            Stopwatch sw = new Stopwatch();
            sw.Start();

            findFingerPoints();

            sw.Stop();
            Console.WriteLine("findFingerPoints:" + sw.ElapsedMilliseconds);
            sw.Restart();

            findFingerLines();

            sw.Stop();
            Console.WriteLine("findFingerLines:" + sw.ElapsedMilliseconds);
            sw.Restart();


            findFingers();

            sw.Stop();
            Console.WriteLine("findFingers:" + sw.ElapsedMilliseconds);

        }

        public List<Finger> getFingers() {
            return fingers;
        }

        private void findFingerPoints() {
            int width = depthImage.getWidth();
            int height = depthImage.getHeight();

            fingerPointsImage = new Image<Gray, byte>(width, height);
            
            for (int y = 0; y < height; y++) {
                for (int x = 0; x < width; x++) {
                    if (edgeImage.isEdgeAt(x, y))
                    {
                        Vector2D startPoint = new Vector2D(x, y);
                        searchFingerPoint(startPoint, new Vector2D(1, 0), fingerPointsImage);
                        searchFingerPoint(startPoint, new Vector2D(0, 1), fingerPointsImage);
                        //searchFingerPoint(startPoint, new Vector2D(1, -1).normalize(), fingerPointsImage);
                        //searchFingerPoint(startPoint, new Vector2D(1, 1).normalize(), fingerPointsImage);
                    }
                }
            }
        }

        private void searchFingerPoint(Vector2D start, Vector2D direction, Image<Gray, byte> fingerPointsImage) {
            int width = fingerPointsImage.Width;
            int height = fingerPointsImage.Height;
            float minFingerSize = 10;
            float maxFingerSize = 30;
            direction = direction.normalize();


            bool edgeFound = false;
            Vector2D edge = start + minFingerSize * direction;
            float index = minFingerSize;
            float maxXIndex = (direction.X == 0) ? float.MaxValue : ((direction.X > 0) ? ((width - start.X) / direction.X) : (-start.X / direction.X));
            float maxYIndex = (direction.Y == 0) ? float.MaxValue : ((direction.Y > 0) ? ((height - start.Y) / direction.Y) : (-start.Y / direction.Y));

            float maxIndex = Math.Min(maxFingerSize, Math.Min(maxXIndex, maxYIndex));

            if (index < maxIndex)
            {
                Int16 depthStart = depthImage.getDepthAt((int)start.X, (int)start.Y);
                Int16 depthEdge = depthImage.getDepthAt((int)edge.X, (int)edge.Y);

                if (depthStart > depthEdge)
                {
                    while (!edgeFound && index < maxIndex)
                    {
                        if (edgeImage.isEdgeAt((int)edge.X, (int)edge.Y))
                            edgeFound = true;
                        else
                        {
                            index++;
                            edge += direction;
                        }
                    }
                }
            }
            
            if (edgeFound && fingerDepthTest(start, edge))
            {
                Vector2D mid = 0.5f * (start + edge);
                fingerPointsImage.Data[(int)mid.Y, (int)mid.X, 0] = byte.MaxValue;
                fingerImage.setFingerAt((int)mid.X, (int)mid.Y, FingerImageState.possibleFinger);
            }
        }

        private bool fingerDepthTest(Vector2D p1, Vector2D p2) {
            Vector2D mid = 0.5f * (p1 + p2);
            Int16 depthP1 = depthImage.getDepthAt((int)p1.X, (int)p1.Y);
            Int16 depthMid = depthImage.getDepthAt((int)mid.X, (int)mid.Y);
            Int16 depthP2 = depthImage.getDepthAt((int)p2.X, (int)p2.Y);
            return (depthP1 > depthMid && depthMid < depthP2);
        }

        private void findFingerLines() {
            float maxCombinableAngle = (float) (Math.PI * 20 / 180);
            float maxCombinableParallelDistance = 10;
            float maxCombinableVerticalDistance = 10;

            fingerPointsImage = fingerPointsImage.Dilate(1);
            LineSegment2D[] lineSegments = fingerPointsImage.HoughLinesBinary(1, Math.PI / 90, 10, 20, 2)[0];

            fingerLines = new List<Line2D>();
            foreach (LineSegment2D lineSegment in lineSegments)
            {
                Line2D line = new Line2D(new Vector2D(lineSegment.P1), new Vector2D(lineSegment.P2));

                List<Line2D> combineableLines = new List<Line2D>();
                foreach(Line2D fingerLine in fingerLines) {
                    float angle = line.getAngleBetween(fingerLine);
                    if (angle <= maxCombinableAngle) {
                        float parallelDistance = line.getParallelDistanceTo(fingerLine);
                        if (parallelDistance <= maxCombinableParallelDistance) {
                            float verticalDistance = line.getVerticalDistanceTo(fingerLine);
                            if (verticalDistance <= maxCombinableVerticalDistance) {
                                combineableLines.Add(fingerLine);
                            }
                        }
                    }
                }
                
                if (combineableLines.Count == 0) {
                    fingerLines.Add(line);
                }
                else {
                    foreach (Line2D mergableLine in combineableLines) {
                        fingerLines.Remove(mergableLine);
                    }
                    combineableLines.Add(line);
                    fingerLines.Add(new Line2D(combineableLines));
                }
            }
        }

        private void findFingers()
        {
            float minLength = 20;
            float minNumCombinedLines = 2;

            fingers = new List<Finger>();
            foreach (Line2D line in fingerLines) {
                if (line.Length >= minLength && line.NumCombinedLines >= minNumCombinedLines)
                {
                    fingers.Add(new Finger(line));
                    fingerImage.drawLine(line, FingerImageState.fingerDetected);
                }
            }

        }
    }
}
